from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from sparkden.models.knowledge import FetchDataResult, KnowledgeDataSource


class BaseDataFetcher(ABC):
    @abstractmethod
    def fetch(self, data_source: "KnowledgeDataSource") -> "FetchDataResult":
        pass



from enum import StrEnum

from litestar.plugins.sqlalchemy import base
from sqlalchemy import ForeignKey, Index, text
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import Mapped, mapped_column


class UserRole(StrEnum):
    ADMIN = "admin"
    USER = "user"


class DataSourceType(StrEnum):
    UPLOAD = "upload"
    API = "api"


class DataSourceStatus(StrEnum):
    PROCESSING = "processing"
    READY = "ready"
    ERROR = "error"


class VectorDistance(StrEnum):
    COSINE = "Cosine"
    EUCLID = "Euclid"
    DOT = "Dot"
    MANHATTAN = "Manhattan"


class RetrievalMode(StrEnum):
    DENSE = "dense"
    SPARSE = "sparse"
    HYBRID = "hybrid"


class User(base.UUIDAuditBase):
    __tablename__ = "users"
    name: Mapped[str]
    username: Mapped[str | None]
    password: Mapped[bytes | None]
    avatar: Mapped[str | None]
    role: Mapped[UserRole] = mapped_column(default=UserRole.USER)
    extra_info: Mapped[dict | None] = mapped_column(JSONB)


class KnowledgeCollection(base.UUIDAuditBase):
    __tablename__ = "knowledge_collections"

    id: Mapped[str] = mapped_column(primary_key=True)
    name: Mapped[str]
    vector_dimensions: Mapped[int] = mapped_column(default=1024)
    vector_distance: Mapped[VectorDistance] = mapped_column(
        default=VectorDistance.COSINE
    )
    retrieval_mode: Mapped[RetrievalMode] = mapped_column(default=RetrievalMode.DENSE)


class KnowledgeDataSource(base.UUIDAuditBase):
    __tablename__ = "knowledge_data_sources"
    __table_args__ = (
        Index("idx_knowledge_data_sources_collection_id", "collection_id"),
        Index(
            "idx_knowledge_data_sources_file_hash", text("(extra_info ->> 'file_hash')")
        ),
    )

    name: Mapped[str]
    definition_id: Mapped[str]
    type: Mapped[DataSourceType]
    status: Mapped[DataSourceStatus]
    extra_info: Mapped[dict | None] = mapped_column(JSONB)
    collection_id: Mapped[str] = mapped_column(ForeignKey("knowledge_collections.id"))



from google.adk.tools import ToolContext
from pydantic import ValidationError

from sparkden.models.shared import BaseModel

from .models import (
    ToolResponse,
    ToolResponseStatus,
    UserApprovalRequest,
)


def update_approval_request_state(
    *,
    resource: dict,
    reason: str,
    tool_call_id: str,
    tool_call_name: str,
    tool_context: ToolContext,
) -> ToolResponse:
    tool_context.state["user_approval_request"] = UserApprovalRequest(
        resource=resource,
        reason=reason,
        tool_call_id=tool_call_id,
        tool_call_name=tool_call_name,
    ).model_dump()
    tool_context.actions.skip_summarization = True
    return ToolResponse(
        result=None,
        status=ToolResponseStatus.PENDING,
    )


def update_approval_result_state(
    *,
    resource: dict | None,
    state_key: str,
    tool_context: ToolContext,
) -> None:
    tool_context.state[state_key] = resource
    tool_context.state["user_approval_request"] = None


def validate_tool_param(
    tool_name: str, param_value: dict, param_type: type[BaseModel]
) -> ToolResponse | None:
    try:
        param_type.model_validate(param_value)
    except ValidationError as e:
        error_str = f"""Invoking `{tool_name}()` failed as the input parameters validation failed:
{str(e)}
You could retry calling this tool, but it is IMPORTANT for you to follow the input parameters schema."""
        return ToolResponse(
            status=ToolResponseStatus.ERROR,
            error=error_str,
        )
    return None



from typing import Annotated

from litestar import Controller, Request, get, post
from litestar.di import Provide
from litestar.exceptions import ValidationException
from litestar.plugins.pydantic import PydanticDTO
from litestar.response import ServerSentEvent
from sparkden.adk.session_service import SessionService
from sparkden.models.shared import BaseModel, OffsetPagination
from sparkden.models.task import CreateTaskParams, SendMessageParams, Task, TaskItem
from sparkden.services.task import TaskService
from sparkden.shared.utils import dto_config, getenv


class GetTaskResponse(BaseModel):
    task: Task | None


CreateTaskParamsDTO = PydanticDTO[Annotated[CreateTaskParams, dto_config()]]
SendMessageParamsDTO = PydanticDTO[Annotated[SendMessageParams, dto_config()]]
TaskDTO = PydanticDTO[Annotated[Task, dto_config()]]
GetTaskResponseDTO = PydanticDTO[Annotated[GetTaskResponse, dto_config()]]
ListTasksResponseDTO = PydanticDTO[Annotated[OffsetPagination[TaskItem], dto_config()]]


def get_task_service(request: Request) -> TaskService:
    return TaskService(
        adk_session_service=SessionService(db_url=getenv("DATABASE_URL")),
        user_id=str(request.user.id),
    )


class TaskController(Controller):
    path = "/tasks"
    dependencies = {
        "task_service": Provide(get_task_service, sync_to_thread=True),
    }

    @get("/", return_dto=ListTasksResponseDTO)
    async def list_tasks(
        self, task_service: TaskService, limit: int = 30, offset: int = 0
    ) -> OffsetPagination[TaskItem]:
        return await task_service.list_tasks(limit=limit, offset=offset)

    @post("/", dto=CreateTaskParamsDTO, return_dto=TaskDTO)
    async def create_task(
        self, task_service: TaskService, data: CreateTaskParams
    ) -> Task:
        return await task_service.create_task(
            title=data.title, assistant_id=data.assistant_id, task_id=data.task_id
        )

    @get("/{task_id:str}", return_dto=GetTaskResponseDTO)
    async def get_task(
        self, task_service: TaskService, task_id: str, assistant_id: str
    ) -> GetTaskResponse:
        task = await task_service.get_task(assistant_id=assistant_id, task_id=task_id)
        return GetTaskResponse(task=task)

    @post("/{task_id:str}/run", dto=SendMessageParamsDTO)
    async def run_task(
        self, task_service: TaskService, task_id: str, data: SendMessageParams
    ) -> ServerSentEvent:
        message = data.message or data.message_part
        if message is None:
            raise ValidationException(
                status_code=400, detail="param message or message_part is required"
            )

        return ServerSentEvent(
            task_service.run_task(
                assistant_id=data.assistant_id, task_id=task_id, message=message
            )
        )



from enum import StrEnum
from typing import Callable, NotRequired, TypedDict

from google.adk.agents import BaseAgent
from litestar import Controller
from pydantic import ConfigDict

from sparkden.models.knowledge import DataSourceType
from sparkden.models.shared import BaseModel, ExtraInfoMixin, base_model_config
from sparkden.services.knowledge.data_loaders.data_fetchers.base import (
    BaseDataFetcher,
)
from sparkden.services.knowledge.data_loaders.data_splitters.base import (
    BaseDataSplitter,
)


class ToolResponseStatus(StrEnum):
    PENDING = "pending"
    SUCCESS = "success"
    ERROR = "error"


class ToolResponse[ResultT](TypedDict):
    """The base tool response."""

    result: NotRequired[ResultT]
    """The result of the tool."""

    status: NotRequired[ToolResponseStatus]
    """The status of the tool response."""

    error: NotRequired[str]
    """The error message of the tool response."""


class UserApprovalRequest[RessourceT](BaseModel):
    """The user approval request."""

    resource: RessourceT
    """The resource requested to be approved."""

    reason: str
    """The reason for the approval request."""

    tool_call_id: str
    """The tool call id."""

    tool_call_name: str
    """The name of the tool call."""


class TaskStatus(StrEnum):
    IDLE = "idle"
    RUNNING = "running"


class ProgressItemStatus(StrEnum):
    COMPLETED = "completed"
    RUNNING = "running"
    PENDING = "pending"


class KnowledgeDataSourceDefinition(ExtraInfoMixin, BaseModel):
    model_config = ConfigDict(
        **base_model_config,
        arbitrary_types_allowed=True,
    )
    id: str
    type: DataSourceType
    data_fetcher: BaseDataFetcher | None = None
    data_splitter: BaseDataSplitter | None = None


class Assistant(BaseModel):
    model_config = ConfigDict(
        **base_model_config,
        arbitrary_types_allowed=True,
    )

    id: str
    disabled: bool = False
    root_agent: BaseAgent
    api_routes: list[type[Controller]] | None = None
    callbacks: dict[str, Callable] | None = None
    data_sources: list[KnowledgeDataSourceDefinition] | None = None
    sequence: int = 9999



from advanced_alchemy.base import metadata_registry
from bcrypt import gensalt, hashpw
from sparkden.shared.pg import get_engine, get_session_maker
from sqlalchemy import exists, select
from sqlalchemy.exc import OperationalError

from .schema import User, UserRole


async def init_db() -> None:
    await seed_db()


async def seed_db() -> None:
    async with get_engine().begin() as conn:
        try:
            await conn.run_sync(metadata_registry.get().create_all)
        except OperationalError as exc:
            print(f"Could not create target metadata.  Reason: {exc}")

    async with get_session_maker().begin() as db_session:
        statement = select(exists().where(User.id.is_not(None)))
        result = await db_session.execute(statement)
        if not result.scalar():
            try:
                for index, user in enumerate(users):
                    is_admin = index in [0, 1]
                    db_session.add(
                        User(
                            name=user,
                            username=f"test{index + 1}",
                            password=hashpw("password123".encode(), gensalt()),
                            extra_info={"college": "计算机学院"},
                            role=UserRole.ADMIN if is_admin else UserRole.USER,
                        )
                    )
                # 事务会在 begin() 上下文管理器退出时自动提交
                print(f"Successfully seeded {len(users)} users")
            except Exception as exc:
                print(f"Error seeding users: {exc}")
                raise


users = [
    "张三",
    "李四",
    "王五",
    "赵六",
    "孙七",
    "周八",
    "吴九",
    "郑十",
    "John Smith",
    "Alice Brown",
]



from typing import Annotated, Any

from bcrypt import checkpw
from litestar import Controller, Request, delete, get, post
from litestar.connection import ASGIConnection
from litestar.exceptions import NotAuthorizedException
from litestar.middleware.session.server_side import (
    ServerSideSessionBackend,
    ServerSideSessionConfig,
)
from litestar.plugins.pydantic import PydanticDTO
from litestar.security.session_auth import SessionAuth
from sparkden.db.schema import User as StorageUser
from sparkden.models.shared import BaseModel
from sparkden.models.user import User
from sparkden.shared.pg import get_session_maker
from sparkden.shared.utils import dto_config
from sqlalchemy import select


class Session(BaseModel):
    user: User


class SignInUser(BaseModel):
    username: str
    password: str


SignInUserDTO = PydanticDTO[Annotated[SignInUser, dto_config()]]
SessionDTO = PydanticDTO[Annotated[Session, dto_config()]]


class SessionController(Controller):
    path = "/sessions"

    @get("/", return_dto=SessionDTO)
    async def get_session(self, request: Request) -> Session:
        return Session(user=request.user)

    @post("/", dto=SignInUserDTO, return_dto=SessionDTO, exclude_from_auth=True)
    async def sign_in(self, data: SignInUser, request: Request) -> Session:
        async with get_session_maker().begin() as db_session:
            user = await db_session.scalar(
                select(StorageUser).where(StorageUser.username == data.username)
            )
            if not user or not user.password:
                raise NotAuthorizedException(
                    status_code=401, detail="Invalid username or password"
                )
            if not checkpw(data.password.encode(), user.password):
                raise NotAuthorizedException(
                    status_code=401, detail="Invalid username or password"
                )
            request.set_session({"user_id": user.id})
            return Session(user=User.from_storage(user))

    @delete("/", exclude_from_auth=True)
    async def sign_out(self, request: Request) -> None:
        request.clear_session()


async def retrieve_user_handler(
    session: dict[str, Any], connection: ASGIConnection
) -> User | None:
    async with get_session_maker().begin() as db_session:
        storage_user = await db_session.get(StorageUser, session["user_id"])
        return User.from_storage(storage_user) if storage_user else None


session_backend_config = ServerSideSessionConfig()

session_auth = SessionAuth[User, ServerSideSessionBackend](
    retrieve_user_handler=retrieve_user_handler,
    session_backend_config=session_backend_config,
    route_handlers=[
        SessionController,
    ],
)



from .base import Distance, Field, FieldCondition, Filter, MatchValue, VectorParams
from .qdrant import QdrantVectorStore

# TODO: determine which vector store to use by app config
VectorStore = QdrantVectorStore

__all__ = [
    "VectorStore",
    "Field",
    "Filter",
    "MatchValue",
    "FieldCondition",
    "Distance",
    "VectorParams",
]



from datetime import datetime
from typing import cast, override

from google.adk.sessions import Session
from google.adk.sessions.base_session_service import ListSessionsResponse
from google.adk.sessions.database_session_service import (
    DatabaseSessionService,
    StorageSession,
)


class SessionService(DatabaseSessionService):
    @override
    async def list_sessions(
        self,
        *,
        app_name: str | None = None,
        user_id: str,
        limit: int = 10,
        offset: int = 0,
    ) -> ListSessionsResponse:
        with self.database_session_factory() as session:
            query = session.query(StorageSession)
            if app_name is not None:
                query = query.filter(StorageSession.app_name == app_name)
            results = (
                query.filter(StorageSession.user_id == user_id)
                .order_by(StorageSession.update_time.desc())
                .offset(offset)
                .limit(limit)
                .all()
            )
            sessions = []
            for storage_session in results:
                session = Session(
                    app_name=storage_session.app_name,
                    user_id=user_id,
                    id=storage_session.id,
                    state=storage_session.state,
                    last_update_time=cast(
                        datetime, storage_session.update_time
                    ).timestamp(),
                )
                sessions.append(session)
            return ListSessionsResponse(sessions=sessions)

    async def sessions_total_count(
        self, *, app_name: str | None = None, user_id: str
    ) -> int:
        with self.database_session_factory() as session:
            query = session.query(StorageSession)
            if app_name is not None:
                query = query.filter(StorageSession.app_name == app_name)
            return query.filter(StorageSession.user_id == user_id).count()



from typing import AsyncGenerator

from google.adk.agents.run_config import RunConfig, StreamingMode
from google.adk.runners import Runner
from google.genai import types
from litestar.response import ServerSentEventMessage
from sparkden.adk.session_service import SessionService
from sparkden.assistants import assistants
from sparkden.models.shared import OffsetPagination

from ..models.task import Task, TaskItem, TaskSSEData


class TaskService:
    def __init__(self, adk_session_service: SessionService, user_id: str):
        self.adk_session_service = adk_session_service
        self.user_id = user_id

    async def list_tasks(
        self, *, limit: int = 30, offset: int = 0
    ) -> OffsetPagination[TaskItem]:
        list_response = await self.adk_session_service.list_sessions(
            user_id=self.user_id, limit=limit, offset=offset
        )
        total_count = await self.adk_session_service.sessions_total_count(
            user_id=self.user_id
        )
        return OffsetPagination(
            items=[
                TaskItem.from_session(session) for session in list_response.sessions
            ],
            total=total_count,
            limit=limit,
            offset=offset,
        )

    async def create_task(
        self, *, title: str, assistant_id: str, task_id: str | None = None
    ) -> Task:
        session = await self.adk_session_service.create_session(
            app_name=assistant_id,
            user_id=self.user_id,
            state={"title": title, "status": "idle"},
            session_id=task_id,
        )
        return Task.from_session(session)

    async def get_task(self, *, assistant_id: str, task_id: str) -> Task | None:
        session = await self.adk_session_service.get_session(
            app_name=assistant_id, user_id=self.user_id, session_id=task_id
        )
        if session is None:
            return None
        return Task.from_session(session)

    async def run_task(
        self, *, assistant_id: str, task_id: str, message: str | types.Part
    ) -> AsyncGenerator[ServerSentEventMessage, None]:
        assistant = assistants.get_assistant(assistant_id)
        if assistant is None:
            raise ValueError(f"Assistant {assistant_id} not found")

        runner = Runner(
            app_name=assistant_id,
            agent=assistant.root_agent,
            session_service=self.adk_session_service,
        )

        new_message = types.Content(
            role="user",
            parts=[
                message if isinstance(message, types.Part) else types.Part(text=message)
            ],
        )

        accumulated_text = ""
        async for event in runner.run_async(
            user_id=self.user_id,
            session_id=task_id,
            new_message=new_message,
            run_config=RunConfig(streaming_mode=StreamingMode.SSE),
        ):
            event_name, event_data = None, None
            if event.partial:
                event_name = "message/partial"
                accumulated_event = event.model_copy(deep=True)
                if (
                    accumulated_event.content
                    and accumulated_event.content.parts
                    and accumulated_event.content.parts[0].text
                ):
                    accumulated_text += accumulated_event.content.parts[0].text
                    accumulated_event.content.parts[0].text = accumulated_text
                # TODO: support partial function call
                event_data = TaskSSEData.from_event(
                    accumulated_event,
                ).model_dump_json(exclude_none=True, by_alias=True)
            else:
                accumulated_text = ""
                event_name = "message"
                event_data = TaskSSEData.from_event(
                    event,
                ).model_dump_json(
                    exclude_none=True,
                    by_alias=True,
                )
            yield ServerSentEventMessage(
                data=event_data,
                event=event_name,
            )



from typing import Any

from google.adk.agents.callback_context import CallbackContext
from google.genai import types

from .models import ProgressItemStatus, ToolResponse


def update_progress_info(
    ctx: CallbackContext,
    key: str,
    status: ProgressItemStatus = ProgressItemStatus.RUNNING,
) -> None:
    progress_info = ctx.state.get("progress_info") or {}
    progress_info[key] = status
    ctx.state["progress_info"] = progress_info


def reset_progress_info(ctx: CallbackContext, key: str | None = None) -> None:
    progress_info = ctx.state.get("progress_info")
    if key and progress_info:
        progress_info.pop(key, None)
        ctx.state["progress_info"] = progress_info
    else:
        ctx.state["progress_info"] = None


def update_state_if_necessary(ctx: CallbackContext, delta: dict[str, Any]) -> None:
    for key, value in delta.items():
        if ctx.state.get(key) != value:
            ctx.state[key] = value


def find_approved_resource_in(content: types.Content | None) -> dict[str, Any] | None:
    if not (content and content.parts):
        return None
    for part in content.parts:
        if part.function_response and part.function_response.response:
            result = ToolResponse[dict[str, Any]](
                **part.function_response.response
            ).get("result")
            if result and result.get("resource"):
                return result.get("resource")

    return None



from datetime import datetime
from typing import Any, Callable
from zoneinfo import ZoneInfo

from google.adk.agents import LlmAgent
from google.adk.agents.callback_context import CallbackContext
from google.adk.models.llm_request import LlmRequest
from google.adk.models.llm_response import LlmResponse
from google.adk.tools import BaseTool, ToolContext

from .helpers import (
    reset_progress_info,
    update_progress_info,
    update_state_if_necessary,
)
from .models import ToolResponse

DEFAULT_TIMEZONE = ZoneInfo("Asia/Shanghai")


class Agent(LlmAgent):
    def __init__(
        self,
        /,
        **data,
    ) -> None:
        # Define callback types and their default methods
        callback_types = {
            "before_agent": self.default_before_agent_callback,
            "after_agent": self.default_after_agent_callback,
            "before_tool": self.default_before_tool_callback,
            "after_tool": self.default_after_tool_callback,
            "before_model": self.default_before_model_callback,
            "after_model": self.default_after_model_callback,
        }

        # Process each callback type
        for callback_type, default_callback in callback_types.items():
            callback_key = f"{callback_type}_callback"
            user_callback = data.pop(callback_key, None)

            # Wrap the callback if provided, otherwise use default
            if user_callback is not None:
                if callback_type.startswith("before"):
                    wrapped_callback = self.wrap_before_callback(
                        default_callback, user_callback
                    )
                elif callback_type.startswith("after"):
                    wrapped_callback = self.wrap_after_callback(
                        default_callback, user_callback
                    )
                else:
                    raise ValueError(f"Invalid callback type: {callback_type}")
            else:
                wrapped_callback = default_callback

            data[callback_key] = wrapped_callback

        super().__init__(**data)

    @staticmethod
    def default_before_agent_callback(
        callback_context: CallbackContext,
    ) -> None:
        update_progress_info(callback_context, callback_context.agent_name)
        update_state_if_necessary(
            callback_context,
            {
                "system_time": datetime.now(tz=DEFAULT_TIMEZONE).isoformat(),
            },
        )
        return None

    @staticmethod
    def default_after_agent_callback(
        callback_context: CallbackContext,
    ) -> None:
        return None

    @staticmethod
    def default_before_tool_callback(
        tool: BaseTool, args: dict[str, Any], tool_context: ToolContext
    ) -> None:
        return None

    @staticmethod
    def default_after_tool_callback(
        tool: BaseTool,
        args: dict[str, Any],
        tool_context: ToolContext,
        tool_response: ToolResponse,
    ) -> None:
        return None

    @staticmethod
    def default_before_model_callback(
        callback_context: CallbackContext, llm_request: LlmRequest
    ) -> None:
        return None

    @staticmethod
    def default_after_model_callback(
        callback_context: CallbackContext, llm_response: LlmResponse
    ) -> None:
        reset_progress_info(callback_context)
        return None

    @staticmethod
    def wrap_after_callback(
        default_callback: Callable, callback: Callable | None
    ) -> Callable:
        def wrapped_callback(*args: Any, **kwargs: Any) -> Any:
            default_callback(*args, **kwargs)
            return callback(*args, **kwargs) if callback else None

        return wrapped_callback

    @staticmethod
    def wrap_before_callback(
        default_callback: Callable, callback: Callable | None
    ) -> Callable:
        def wrapped_callback(*args: Any, **kwargs: Any) -> Any:
            result = callback(*args, **kwargs) if callback else None
            if result is None:
                default_callback(*args, **kwargs)
            return result

        return wrapped_callback
